/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "app_uart.h"
#include "app_error.h"
#include "nrf_delay.h"
#include "nrf.h"
#include "bsp.h"
#include "micro_esb.h"
#include "uesb_error_codes.h"
#include "nrf_gpio.h"
#include "app_scheduler.h"


#define APP_SCHED_QUEUE_SIZE 10
#define APP_SCHED_EVENT_SIZE sizeof(uesb_payload_t)
	
#define UESB_PIPE0_ADDR {0x12, 0x34, 0x56, 0x78, 0x9A}
#define UESB_PIPE1_ADDR {0xBC, 0xDE, 0xF0, 0x12, 0x23}
#define UESB_PIPE2_ADDR 0x66

#define UESB_TX_PIPE 0

#define MAX_TEST_DATA_BYTES     (15U)                /**< max number of test bytes to be used for tx and rx. */
#define UART_TX_BUF_SIZE 256                         /**< UART TX buffer size. */
#define UART_RX_BUF_SIZE 1                           /**< UART RX buffer size. */

static void uart_print_evt_handler(void * p_event_data, uint16_t event_size)
{
    uint8_t * char_ptr;
    
    APP_ERROR_CHECK_BOOL(event_size != 0);
    
    char_ptr = (uint8_t *) p_event_data;
    
    for (int i = 0; i < event_size; ++i)
    {
        app_uart_put(char_ptr[i]);
    }
}

static void uesb_send_pkt_evt_handler(void * p_event_data, uint16_t event_size)
{
    uesb_payload_t payload;
    
    APP_ERROR_CHECK_BOOL(event_size <= UESB_CORE_MAX_PAYLOAD_LENGTH);
    
    payload.pipe  = UESB_TX_PIPE;
    payload.length = event_size;
    memcpy(payload.data, p_event_data, event_size);
    
    // Stop RX
    uesb_stop_rx(); // Ignore possible UESB_ERROR_NOT_IN_RX_MODE return code
    
    uesb_write_tx_payload(&payload); // Ignore possible errors
}

static void uesb_start_rx_evt_handler(void * p_event_data, uint16_t event_size)
{
    uint32_t err_code;
    
    err_code = uesb_start_rx();
    if (err_code != UESB_SUCCESS)
    {
        // Retry operation
        app_sched_event_put(0, 0, uesb_start_rx_evt_handler);
    }
}  

void uart_error_handle(app_uart_evt_t * p_event)
{
    if (p_event->evt_type == APP_UART_COMMUNICATION_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_communication);
    }
    else if (p_event->evt_type == APP_UART_FIFO_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_code);
    }
		else if (p_event->evt_type == APP_UART_DATA_READY)
		{
				uint8_t rxd_char;
			  uint32_t err_code;
				app_uart_get(&rxd_char);
			  err_code = app_sched_event_put(&rxd_char, 1, uesb_send_pkt_evt_handler);
				APP_ERROR_CHECK(err_code);			
		}
}

void uesb_event_handler()
{
    static uint32_t rf_interrupts;
    static uint32_t tx_attempts; // Retries attempted by ESB
    
    uesb_get_clear_interrupts(&rf_interrupts);
    
    if(rf_interrupts & UESB_INT_TX_SUCCESS_MSK)
    {   
        uint32_t err_code;
        
        // Data successfully sent. Start RX again
        err_code = app_sched_event_put(0, 0, uesb_start_rx_evt_handler);
        APP_ERROR_CHECK(err_code);
    }
    
    if(rf_interrupts & UESB_INT_TX_FAILED_MSK)
    {
        // Data failed to be sent: TODO: try again later
        uesb_flush_tx();
    }
    
    if(rf_interrupts & UESB_INT_RX_DR_MSK)
    {
        uint32_t       err_code;
        uesb_payload_t payload;
        
        // Data has been received: schedule printout
        uesb_read_rx_payload(&payload);
        
        err_code = app_sched_event_put(payload.data, payload.length, uart_print_evt_handler);
        APP_ERROR_CHECK(err_code);
    }
    
    uesb_get_tx_attempts(&tx_attempts);
}

static void uesb_config(void)
{
    uint8_t rx_addr_p0[] = UESB_PIPE0_ADDR;
    uint8_t rx_addr_p1[] = UESB_PIPE1_ADDR;
    uint8_t rx_addr_p2   = UESB_PIPE2_ADDR;
    
    uesb_config_t uesb_config       = UESB_DEFAULT_CONFIG;
    uesb_config.rf_channel          = 5;
    uesb_config.crc                 = UESB_CRC_16BIT;
		uesb_config.dynamic_ack_enabled = 0;
    uesb_config.retransmit_count    = 10;
    uesb_config.retransmit_delay    = 500;
   
    uesb_config.protocol            = UESB_PROTOCOL_ESB_DPL;
    uesb_config.bitrate             = UESB_BITRATE_2MBPS;
    uesb_config.event_handler       = uesb_event_handler;

    uesb_init(&uesb_config);

    uesb_set_address(UESB_ADDRESS_PIPE0, rx_addr_p0);
    uesb_set_address(UESB_ADDRESS_PIPE1, rx_addr_p1);
    uesb_set_address(UESB_ADDRESS_PIPE2, &rx_addr_p2);

}

static void uart_config(void)
{		
	  uint32_t err_code;
    const app_uart_comm_params_t comm_params =
      {
          RX_PIN_NUMBER,
          TX_PIN_NUMBER,
          RTS_PIN_NUMBER,
          CTS_PIN_NUMBER,
          APP_UART_FLOW_CONTROL_DISABLED,
          false,
          UART_BAUDRATE_BAUDRATE_Baud38400
      };
			
		APP_UART_FIFO_INIT(&comm_params,
						 UART_RX_BUF_SIZE,
						 UART_TX_BUF_SIZE,
						 uart_error_handle,
						 APP_IRQ_PRIORITY_LOW,
						 err_code);

    APP_ERROR_CHECK(err_code);
}
int main(void)
{		

    NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
    NRF_CLOCK->TASKS_HFCLKSTART = 1;
    while(NRF_CLOCK->EVENTS_HFCLKSTARTED == 0);		

    // Initialize the scheduler module
    APP_SCHED_INIT(APP_SCHED_EVENT_SIZE, APP_SCHED_QUEUE_SIZE);			
		
		// Configure micro ESB
    uesb_config();			
		// Configure UART
    uart_config();
			
		// Start listening for incoming transmissions
    uesb_start_rx();    
	
    while (true)
    {   
			 app_sched_execute();
        
        // Let the CPU sleep until the next interrupt comes along
        __WFE();
        __SEV();
        __WFE();
    }
}
